/* A demo of edf-based execution of taks.
 *
 * Compilation:
 * cc -lrt rt-edf-demo.c -o rt-edf-demo && ./rt-edf-demo
 *
 * We do not perform all the memory tweaking (e.g., page locking). See
 * rt-nanosleep-demo instead.
 */

#include <inttypes.h>
#include <linux/sched.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/syscall.h>
#include <unistd.h>


// Select one of the three
/*#define MODEL_CBS*/
/*#define MODEL_SPORADIC*/
#define MODEL_PERIODIC


/** The attribute structure for sched_setattr() and sched_getattr().
 *
 * The structure is not provided by glibc. See man sched_setattr() for its
 * definition. */
struct sched_attr {
    uint32_t size;

    uint32_t sched_policy;
    uint64_t sched_flags;

    /* SCHED_NORMAL, SCHED_BATCH */
    int32_t sched_nice;

    /* SCHED_FIFO, SCHED_RR */
    uint32_t sched_priority;

    /* SCHED_DEADLINE (nsec) */
    uint64_t sched_runtime;
    uint64_t sched_deadline;
    uint64_t sched_period;
};


/** A wrapper for the sched_setattr() syscall.
 *
 * This wrapper is not provided by glibc. See man sched_setattr for its
 * definition. */
int sched_setattr(pid_t pid, const struct sched_attr *attr, unsigned int flags)
{
   return syscall(__NR_sched_setattr, pid, attr, flags);
}


#if defined(MODEL_CBS)

/* Model: Aperiodic, bandwidth-guaranteed computing.  */
static void work() {
    /* Just compute, without blocking or calls or yield. The CBS will throttle
     * the thread after 20ms per period. */
    while(1);
}

#elif defined(MODEL_SPORADIC)

static void blocking_wait_for_event() {
    /* This is a poor man's example to wait for some "event". */
    usleep(50000);
}

static void sporadic_job() {
    static unsigned cnt = 0;
    ++cnt;

    if (cnt % 10 == 0)
        putchar('|');
    else
        putchar('-');

    fflush(stdout);
}

/* Model: Sporadic event-handling task. */
static void work() {
    while(1) {
        /* Sleep until next event. */
        blocking_wait_for_event();
        /* After wakeup: React on event, e.g., process data, produce result. */
        sporadic_job();
    }
}

#elif defined(MODEL_PERIODIC)

static void periodic_job() {
    static unsigned cnt = 0;
    ++cnt;

    if (cnt % 10 == 0)
        putchar(':');
    else
        putchar('.');

    fflush(stdout);
}

/* Model: Periodic tasks */
static void work() {
    while(1) {
        /* Cylic work. WCET of 20ms. */
        periodic_job();
        /* Notify scheduler about end of job. Only for real-time scheduling! */
        sched_yield();
    }
}

#endif


int main (int argc, char **argv) {

    /* A task that runs 20ms within 100ms periods. */
    struct sched_attr attr;
    memset(&attr, 0, sizeof(attr));
    attr.size = sizeof(attr);
    attr.sched_policy = SCHED_DEADLINE;
    attr.sched_runtime = 20000000;
    attr.sched_deadline = attr.sched_period = 100000000;

    int flags = 0;
    if (sched_setattr(0, &attr, flags) < 0) {
        perror("sched_setattr() failed");
        return EXIT_FAILURE;
    }

    /* The actual work... */
    work();
    return EXIT_SUCCESS;
}
