/* A demo of nanosleep-based periodic execution of taks.
 *
 * Compilation:
 * cc -lrt rt-nanosleep-demo.c -o rt-nanosleep-demo && ./rt-nanosleep-demo
 *
 * Further information:
 * [1] Marta Rybczyńska, A realtime developer's checklist,
 *     https://lwn.net/Articles/837019/
 * [2] John Ogness, A Checklist for Writing Linux Real-Time Applications,
 *     https://ogness.net/ese2020/ese2020_johnogness_rtchecklist.pdf
 */

#include <inttypes.h>
#include <linux/sched.h>
#include <malloc.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <time.h>
#include <unistd.h>

#define MSEC_TO_NSEC(x) (x * 1000000ull)
#define NSEC_PER_SEC 1000000000ull


/** Normalize timespec struc such that nsec is less then NSEC_PER_SEC. */
static inline void normalize_ts(struct timespec* tv) {
    tv->tv_sec += tv->tv_nsec / NSEC_PER_SEC;
    tv->tv_nsec %=  NSEC_PER_SEC;
}


static void work() {
    static unsigned cnt = 0;
    ++cnt;

    if (cnt % 10 == 0)
        putchar(':');
    else
        putchar('.');

    fflush(stdout);
}


int main(int argc, char *argv[]) {

    /* Set scheduler to SCHED_FIFO. Unlike for SCHED_DEADLINE, we do not need
     * sched_setattr(), but sched_setscheduler() suffices. */
    struct sched_param sp = {
        /* Priority is between 1 (low) and 99 (high). See man sched for details. */
        .sched_priority = 10
    };
    if (sched_setscheduler(0, SCHED_FIFO, &sp) != 0) {
        perror("sched_setscheduler() failed");
        return EXIT_FAILURE;
    }

    /* Priority inheritence is not activated by default, so activate it.
     * However, since this demo is without depdendent threads/processes, we do
     * not. */
    /*pthread_mutexattr_setprotocol(&mattr, PTHREAD_PRIO_INHERIT);*/

    /* Turn off mmap()-based malloc, see [1]. */
    mallopt(M_MMAP_MAX, 0);
    /* Do not trim back heap size, see [1]. */
    mallopt(M_TRIM_THRESHOLD,-1);
    /* Lock memory pages, see [1]. */
    mlockall(MCL_CURRENT | MCL_FUTURE);

    /* Now we could do pre-faulting: Allocating all the memory we will need and
     * access them, i.e., perform a non-zero write, to provoke allocation.
     * Should also be done for the stack pages. We will skip it here.
     *
     * Also, we shall set processor affinity as in [2] in order to avoid cache
     * invalidation when migrating to a different core/processor. */

    /* Start of cyclic execution. If you want to delay the first cycle then add
     * the timeshift to tv before the while loop. */
    struct timespec tv;
    clock_gettime(CLOCK_MONOTONIC, &tv);

    while (1) {
        /* Wait until we reached next 100ms period. Note that we compute
         * absolute points in time, so the runtime of work() does not count. */
        tv.tv_nsec += MSEC_TO_NSEC(100);
        normalize_ts(&tv);
        clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &tv, NULL);

        /* The work we perform */
        work();
    };

	return EXIT_SUCCESS;
}

